/*                   HUpars                                        */
/* kontextfreier Parser in Anlehnung an von Prof. K.Brockhaus      */
/* Autor: Dr. E.Huckert 12/85                                      */
/*        Ringofenstr. 23                                          */
/*        63773 Goldbach/Germany                                   */
/*                                                                 */
/* Diese Version beinhaltet:                                       */
/*  a) den Kernparser                                              */
/*  b) eine einfache Endungsmorphologie                            */
/*  b) Regelauswahl ueber Hashing                                  */
/*                                                                 */
/* Es wird erwartet, dass Regeln und Lexikon syntaktisch korrekt   */
/* und aufsteigend sortiert sind. Eine Syntaxpruefung findet nicht */
/* statt.                                                          */
/* Zur Syntaxpruefung sollte das Programm syncheck vor Starten     */
/* des Parsers verwendet werden.                                   */
/* Version 12/86: Sowohl im Lexikon als auch in den Regeln koennen */
/*                Semanteme eingesetzt werden                      */
/*                                                                 */ 
/* Version fuer IBM - PC und VAX/VMS  mit VT100 aehnlicher         */
/* Terminalschnittstelle                                           */
/* Bei Anpassung an andere Terminals: Unterprogramme curpos und    */
/*                                    clsc ersetzen                */
/* Bei Portierung auf UNIX sollte das Terminal - IO angepasst      */
/* werden                                                          */
/*                                                                 */ 
/* Aufruf: hupars [-t -a axiom ]                                   */
/*         -t : Trace wird eingeschaltet                           */
/*         -a : Axiom folgt (Standardaxiom ist SATZ)               */ 
/* 3.12.86 : Semanteme eingefuehrt                                 */
/* 15.12.86: db_ Routinen geschrieben zur Sythese von DB-Anfragen  */ 
/*           (ausgelagert in husynth.c)                            */
/* 28.12.86: lex_analysis modifiziert (Zahlen + Strings)           */

/* 20-12-1988: tested with TurboC V2.0 compiler on IBM-PC/AT       */
/* 26-12-1988: brackets notation generated                         */
/*             messages now translatable (in message file)         */
/*             tested with VAXC and Microsoft C V4.0               */
/* 1.1.1989:   added some English comments                         */
/*             slight changes in output options                    */
/*             modification in read_sentence                       */
/* 17.7.2002   Parameter decl. changed to new C ANSI style         */
/*             Order or procedures rearranged                      */
/*             Drawing of bracket structures reorganized           */
/* 08.8.2002   sub categories option. added to bracketed output    */
/* 08.9.2002   DB related routines eliminated                      */
/*             New version of tree drawing routines added          */ 
/*             Options -tree and -list and -subcat added           */
/*             Option -m (mode) eliminated                         */
/* 20.11.2015  stylistic enhancements, printfs revised             */
    
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/timeb.h>
#include <time.h>


#define CHAR unsigned char

#define LF               10
#define CR               13
#define BELL             7
#define DICT_SEP	(CHAR)'='

// message numbers
#define MSG_GREAD          3
#define MSG_LREAD          4
#define MSG_EREAD          5
#define MSG_LEXNOTFOUND    6
#define MSG_NEXTSENTENCE   7
#define MSG_BRACKNOT       9
#define MSG_NOTANALYZED    10
#define MSG_RETURN         11


/* Option Analyseliste ausgeben */
#define OPTANALIST   0x04
/* Option Ausgabe in Klammernotation */
#define OPTBRACKETS  0x08
/* Option Ausgabe als Baumstruktur */
#define OPTTREE      0x10
// Option Ausgabe von UK-Werten
#define OPTSUBCAT    0x20

/* Laenge Hashtabelle           */
#define LHSH 32
/* Laenge  einer Hauptkategorie */
#define LAHK  4
/* Maximallaenge eines Semantems */
#define LSEM  20

/* Struktur der Unterkategorienwerte */
typedef struct scvalue
{
  CHAR length;
  CHAR values[1];
}   SCVALUE;

/* Kategorialrahmen */
typedef struct cframe
{
  struct cframe *next;
  CHAR           length;
  CHAR           cstring[1];
}   CFRAME;
  
/* Aufbau eines Lexikoneintrags */
/* swo kann beliebig lang sein  */
/* structure of a lexical entry */
typedef struct lexent
{
  struct lexent *next;       /* Pointer auf Nachfolger           */ 
  SCVALUE       *scvalue;    /* Pointer auf Unterkategorienwerte */
  CFRAME        *cframe;     /* Pointer auf Kategorialrahmen     */
  CHAR          semant[LSEM+2]; /* Semantemstring                */
  CHAR          length;      /* Laenge lex. String               */
  CHAR          swo[1];      /* lexik. String                    */ 
}   LEXENT;

/* Aufbau einer komplexen Kategorie */
/* structure of a complex category  */
typedef struct categ
{
  /* Pointer auf Kategorialrahmen */
  CFRAME  *pcfr;
  /* Pointer auf UK - Wertestring */
  SCVALUE *pscv;
}   CATEG;

/* Aufbau einer Regel */
/* structure of a rule */
typedef struct rule
{
  struct rule *next;            /* Pointer auf Nachfolger   */
  CATEG       rpart[3];
  SCVALUE     *rupbd;           /* Pointer auf Bedingungen */
  CHAR        semant[LSEM+2];   /* Semantem                */  
  int         number;           /* Nummer der Regel        */
}  RULE;

/* Knoten im Analysebaum */
/* a node in the parsing list */
typedef struct node
{
  struct node *next;         /* Pointer auf Nachfolger        */ 
  CFRAME      *pkar;         /* Pointer auf Kategorialrahmen  */
  SCVALUE     *pukw;         /* Pointer auf UK - String       */
  struct node *plko;         /* Pointer auf linke Komponente  */
  struct node *prko;         /* Pointer auf rechte Komponente */
  RULE        *prul;         /* Pointer auf eine Regel        */  
  char        semant[LSEM+2];/* Semantem                      */  
  int         lseg;          /* linke Segmentgrenze           */
  int         rseg;          /* rechte Segmentgrenze          */
} NODE;

/* Aufbau der Datenbankdefinitionen */
/* wird fuer die Synthese gebraucht */
typedef struct dbdefs
{
  CHAR          name[80];       /* Name einer Definition */
  CHAR          kat[20];        /* Kategorie             */
  struct dbdefs *next;          /* Nachfolgerpointer     */ 
}   DBDEFS;

/* structure of a message entry */
typedef struct message
{
  int number;                    /* message number */
  struct message *next;          /* pointer to successor */
  CHAR string[1];                /* 1 is a dummy length */
}   MESSAGE;

/* Regelhash */
/* rules hash */
CHAR    *rhsh[LHSH];

// Woerterbuchhash  - lexicon hash  
LEXENT  *whsh[LHSH];
// Endungshash  - endings hash 
LEXENT  *ehsh[LHSH];

/* global variables for HUpars */
CHAR    s_end_chars[40],w_end_chars[40];
CHAR    AXIOM[32];
CHAR    *bracket_buf;
CHAR    *br_addr;
SCVALUE *hukstr;
NODE    *left_comp, *right_comp;
CHAR    sentence[400];
CHAR    *sadr;
int     trace = 0;
int     no_sentences = 0;
CHAR    sembuf[200];
MESSAGE *start_messages = NULL;
int     anzw, zanw;
int     axp  = 1;
int     ayp  = 1;
int     ymax = 1;
int     anzel, seg_start, seg_end; 
int     maxseg;
int     lsem;
int     mode        = 0;
int     interactive = 1;
unsigned no_drawn_nodes = 0;
NODE     *drawn_nodes[128];

NODE   *start_nodes; /* Adr. Knotenspeicher */
/* Variablen fuer Regelverarbeitung */
RULE   *rul_start,*prev_rule,*last_rule;
/* Variablen fuer Lexikonzugriff */
LEXENT *lex_start,*llast;
/* Variablen fuer morph. Analyse */
LEXENT *end_start;
/* Pointer auf Hauptkategorien */
CFRAME *start_categories,*last_category;   

/* ------------------------------------------------------------ */
// Vorsicht: liefert Wert in ms, nicht sec.!
long get_epoch_time_ms(void)
{
time_t   ltime;
struct timeb tb;
  ltime = time(0);
  ftime(&tb);
  return (long) (ltime * 1000 + (int)tb.millitm);
}   // end UTIL::get_epoch_time_ms()

/* String (Laenge angegeben) ausgeben */
void put_string(CHAR *adr,
                int   l)
{
  while (l-- > 0)
   fputc(*adr++,stdout);
}   /* end put_string */

/* same as put_string, but result in string dest */
void make_string(CHAR *dest,
                 CHAR *adr,
                 int   l)
{
  while (l-- > 0)
   *dest++ = *adr++;
  *dest = 0;
}   /* end make_string */

/* String compare Laenge lan    */
/* out: 0 bei Identitaet        */
/*     >0 wenn s1 > s2          */
/*     <0 wenn s1 < s2          */
/* nicht durch strncmp ersetzen */
/* don't replace by strncmp     */
int my_strcmp(CHAR *s1,
              CHAR *s2,
              int   lan)
{
int i;
  while (lan-- > 0)
  {
    i = *s1++ - *s2++;
    if (i != 0) return(i);
  }
  return(0);
}   /* end my_strcmp */

// -----------------------------------------------------------
/* string copy. don't replace by strncpy! */
void my_strncpy(CHAR *dest,
                CHAR *src,
                int   lan)
{
  while (lan-- > 0)
    *dest++ = *src++;
}   /* end my_strncpy */

// -----------------------------------------------------------
/* Zeichen in Grossbuchstaben konvertieren */
/* May not be used for German Umlauts!     */
CHAR my_toupper(CHAR c)
{
  if ((c <= 'z') && (c >= 'a'))
  {
    c = toupper(c);
    return(c);
  }
  if ((c >= '{') && (c <= '~'))
     return(c - 0x20);
  return(c);
}   /* end my_toupper() */

// -----------------------------------------------------------
/* print a message via a message number */
/* returns -1 if error                  */
int hup_message(int mno)
{
MESSAGE *hadr;
  hadr = start_messages;
  while (hadr != NULL)
  {
    if (hadr->number == mno)
    {
      int n, m;
      char buf[256];
      m = 0;
      for (n=0; hadr->string[n] != 0; n++)
      {
        if (hadr->string[n] == '\\')
        {
          if (hadr->string[n+1] == 'n')
            buf[m++] = '\n';
          if (hadr->string[n+1] == 'r')
            buf[m++] = '\r';
          n++;
          continue;
        }
        buf[m++] = hadr->string[n];
      }
      buf[m] = 0;
      printf("%s",buf);
      return(0);
    }
    hadr = hadr->next;
  }
  return(-1);
}   /* end hup_message() */

// -----------------------------------------------------------
/* Aufforderung zum Weitermachen */
// Returns: 0 if no further input
int enter_return()
{
  char inpStr[64];
  //
  hup_message(MSG_RETURN);
  if (fgets(inpStr, sizeof(inpStr), stdin) == NULL)
    return 0;
  if (inpStr[0] != 0) 
    return (toupper(inpStr[0]));
  return 0;
}   /* end enter_return() */

// -----------------------------------------------------------
/* Wortlaut diad mit Suchwort swad vergleichen */
LEXENT *compare_word(LEXENT *diad,  // word in dictionary 
                     CHAR   *swad,
                     int     l)     // required comparison length
{
int m;
  /* Laengen vergleichen */
  if (diad->length == l)
  {
    /* Wortlaut vergleichen */
    for (m=0; m < l; m++) 
    {
      if (diad->swo[m] != swad[m]) 
        return(NULL);
    }
    return(diad);
  }
  return(NULL);
}   /* end compare_word() */

// -----------------------------------------------------------
/* Meldungen ausgeben mit exit ! */
void errexit(char *message)
{
  printf("\n\t\t------ %s ------\n", message);
  exit(1);
}   /* end errexit */

// -----------------------------------------------------------
// Print the subcategories for a single node 
void print_subcategories(SCVALUE *pukw)
{
int i, j, ll, flag;
  if (pukw == NULL)
    return;
  //
  for (i=0; i < pukw->length; i++)
  {
     if (pukw->values[i] == 0xff)
     {
       printf(" ()");
     }
     else
     {
       printf(" (");
       ll = pukw->values[i];
       flag = 0;
       for (j=8; j > 0; j--)
       {
         if (ll & 0x80)
         {
           if (flag)
           {
             fputc(',',stdout);
           } 
           printf("%d",j);
           flag = 1;
         }
         ll = ll << 1;
       }             
       printf(")");
     }   // else
  }  // for i...
}   // end print_subcategories()
 
// -----------------------------------------------------------
// Print all subcategories for the nodes collected in array drawn_nodes
print_all_subcategories()
{
int n;
  //printf("\n");
  for (n=0; n < (int)no_drawn_nodes; n++)
  {
     printf(":%d ", n + 1);
     print_subcategories(drawn_nodes[n]->pukw);
     printf("\n");
  }
}   // end print_all_subcategories()

// -----------------------------------------------------------
/* Return the node address for the axiom node */
/* May return NULL!                           */
NODE *find_axiom_node(NODE *start, 
                      CHAR *axiom)
{
int   lax;
NODE *pnode;
NODE *paxiom;
  paxiom  = NULL;
  pnode   = start;
  lax     = strlen((char *)axiom);
  while (pnode != NULL)
  {
    if (my_strcmp(pnode->pkar->cstring, axiom, lax) == 0)
    {
      if ((pnode->rseg == maxseg) && (pnode->lseg == 1))
      { 
        paxiom = pnode;
        break;
      }
    }
    pnode = pnode->next;
  }   /* end while ... */
  if (trace > 1)
  {
    printf("find_axiom_node(): %p\n", paxiom);
    fflush(stdout);
  }
  return paxiom;
}   // end find_axiom_node() */

// -----------------------------------------------------------
/* generate output in bracket notation                          */
/* This procedure is recursive                                  */
/* Uses external address br_addr to store the bracket structure */
/* Uses external array drawn_nodes to keep the addresses of     */
/*     drawn nodes. Uses ext. var. no_drawn_nodes to count them */
void draw_brackets(NODE *start)
{
CFRAME *ha;
int i;
  if (trace)
  {
    fprintf(stdout, "in draw_brackets()\n");
    fflush(stdout);
  }
  if (start == NULL)
  {
    fprintf(stdout, "draw_brackets(): no start node specified\n");
    fflush(stdout);
    return;
  }
  /* insert label */
  ha = start->pkar;
  if (ha == NULL)
  {
    fprintf(stdout,  "draw_brackets(): internal error 1\n");
    fflush(stdout);
    return;
  }
  drawn_nodes[no_drawn_nodes++] = start;
  //strncpy((char *)br_addr, (char *)(ha->cstring), ha->length);
  memcpy((char *)br_addr, (char *)(ha->cstring), ha->length);
  br_addr += ha->length;
  //for (i = 0; i < ha->length; i++)
  //  *br_addr++ = ha->cstring[i];
  if (mode & OPTSUBCAT)
  {
    sprintf((char *)br_addr, ":%d", no_drawn_nodes);
    i = 2;
    if (no_drawn_nodes > 9)  i++;
    if (no_drawn_nodes > 99) i++;
    br_addr += i;
  }

  /* climb down the left branch */
  if (start->plko == NULL)
    return;
  *br_addr++ = '(';
  draw_brackets(start->plko);
  if (start->prko != NULL)
  {
    /* climb down the right branch */
    if (*(br_addr-1) != ')') *br_addr++ = ' ';
    draw_brackets(start->prko);
  }
  *br_addr++ = ')';
}   /* end draw_brackets() */

#define DIM1 64
#define DIM2 256
#define DIM3 8

// -----------------------------------------------------------
// draw an analysis tree from a bracketed structure
// Assumes that the axion has been reached
void draw_analysis_tree(char *brack)
{
int  ypos = 0;
int  xpos = 0;
int  ymax = 0;
int  xmax = 0;
int  ll, m, n;
char *caddr;
char *buf;
char ch;
short stack[DIM1];      // keeps X positions of predecessors
short stackpos = 0;     // position in the stack
  //
  if (trace)
  {
    fprintf(stdout, "in draw_analysis_tree()\n");
    fflush(stdout);
  }
  //
  buf   = (char *)malloc(DIM1 * DIM2);
  memset(buf, ' ', DIM1 * DIM2);
  for (n=0; brack[n] != 0; n++)
  {
    ch = brack[n];
    //printf("ch=%c xpos=%d ypos=%d xmax=%d ymax=%d\n", 
    //        ch, xpos, ypos, xmax, ymax);
    if (ch == '<') continue;
    if (ch == '>') continue;
    if (ch == '(')
    {
      ypos += 3;
      if (ypos >= DIM1) break;
      if (ypos > ymax) ymax = ypos;
      stackpos++;
      if (stackpos >= DIM1) stackpos = DIM1 - 1;
      continue;
    }
    if (ch == ')')
    {
      ypos -= 3;
      if (brack[n+1] != ')')
      {
        xpos++;
        if ((xpos * DIM3) >= (DIM2 - DIM3)) break;
        if (xpos > xmax) xmax = xpos;
      }
      stackpos--;
      continue;
    }
    if (ch == ' ')
    {
      xpos++; 
      if ((xpos * DIM3) >= (DIM2 - DIM3)) break;
      if (xpos > xmax) xmax = xpos;
      continue;
    }
    //
    // must be a category here
    stack[stackpos] = xpos;
    ll              = 0;
    caddr           = buf + (ypos * DIM2) + (DIM3 * xpos);
    for (m=n; brack[m] != 0; m++)
    {
      if (strchr("()> ", brack[m]) != NULL)
        break;
      ll++;
      *caddr++ = brack[m];
    }
    if (ypos > 0)
    {
      int pred_xpos;
      // vertical line to the predecessor
      for (m=1; m <= 2; m++)
      {
        caddr = buf + ((ypos - m) * DIM2) + (xpos * DIM3);
        *caddr = '|';
      }
      //
      // horizontal line to the predecessor
      pred_xpos = (stack[stackpos - 1] * DIM3) + 3;
      caddr = buf + ((ypos - 3) * DIM2);
      for (m = (xpos * DIM3); m > pred_xpos; m--)
      {
        caddr[m] = '_';
      } 
    }
    n += (ll - 1);
  }   // for n...
  //
  // output the tree
  caddr = buf;
  for (n=0; n <= ymax; n++)
  {
    caddr[xmax * DIM3] = '\0'; 
    fprintf(stdout, "%s\n", caddr);
    fflush(stdout);
    caddr += DIM2;
  } 
  //
  free(buf);
  //printf("xmax=%d\n", xmax);
}   // end draw_analysis_tree()

// -----------------------------------------------------------
/* Nachfolgereintraege(Knotensp.) setzen */
/* set link node entries                 */
void set_predecessor(NODE *firstnod,
                     NODE *anod)
{
NODE *ha,*last;
  /* Nachfolgereintrag im neuen Knoten */
  anod->next = NULL;
  if (firstnod == anod) return;
  /* Nachfolgereintrag im Vorgaenger */
  ha = firstnod;
  while (ha != NULL)
  {
    last = ha;
    ha = ha->next;
  }
  last->next = anod;
} /* end set_predecessor */

// -----------------------------------------------------------
/* Knoten im Kateg.speicher ablegen   */
/* insert a node in the analysis list */
void keep_node(CFRAME *kap,
               SCVALUE *ukp,
               NODE *left_comp,
               NODE *right_comp,
               RULE *renu,
               int  sgma,
               int  sgme,
               char *semant)
{
NODE *anf;
  anf = (NODE *)malloc(sizeof(NODE));
  if (anf == NULL)
    errexit("keep_node: malloc Fehler");
  if (start_nodes == NULL) start_nodes = anf;
  /* Kategorienpointer und UK-Pointer */
  anf->pkar = kap;
  anf->pukw = ukp;
  /* Pointer auf linke Kompon. */
  anf->plko = left_comp;
  /* Pointer auf rechte Kompon. */
  anf->prko = right_comp;
  /* Pointer auf Regel */
  anf->prul = renu;
  /* Seg.grenzen (Anfang, Ende) */
  anf->lseg = sgma;
  anf->rseg = sgme; 
  /* Semantem */
  strcpy(anf->semant,semant);
  /* Nachfolgepointer */
  set_predecessor(start_nodes,anf);
}   /* end keep_node() */

// -----------------------------------------------------------
/* Prueft, ob Regel auf gleiche Knoten schon       */
/* angewendet wurde                                */
/* Returnwerte: 0 falls noch nicht angewendet      */
/* Check if the same rule has been applied already */
int already_applied(RULE *rn,
                    NODE *lko,
                    NODE *rko)
{
NODE *ha;
  ha = start_nodes;
  while (ha != NULL)
  {
    if (ha->plko == lko)
    {
      if (ha->prko == rko)
      {
        if (ha->prul == rn) return(1);
      }
    }
    ha = ha->next;
  }
  return(0);
}   /* end already_applied */

// -----------------------------------------------------------
/* Neuen Knoten anlegen, falls gleiche Regel           */
/* auf gleiche Komponenten nicht schon angewendet      */
/* In hukstr muss der String der neuen UK-Werte stehen */
/* Generate a new node under some conditions           */
void make_node(RULE *ra)
{
  SCVALUE *uanf;
  CFRAME *hkr3;
  //
  /* pruefen, ob Regel schon angewendet */ 
  if (already_applied(ra, left_comp, right_comp) == 0)
  {
    /* gleiche Regel wurde noch nicht angewendet */
    zanw++;
    hkr3 = ra->rpart[2].pcfr;
    if (trace)
    {
      printf("\n\t\t\tmake_node:hkr3=%p ", hkr3);
      put_string(hkr3->cstring,hkr3->length);
    }
    /* Neuen UK-Wert-String anlegen */
    uanf = NULL;
    if (hukstr->length != 0)
    {
      uanf = (SCVALUE *)malloc(sizeof(SCVALUE) + hukstr->length);
      if (uanf == NULL)
        errexit("make_node: malloc Fehler");
      uanf->length = hukstr->length;
      my_strncpy(uanf->values,hukstr->values,hukstr->length);
    }
    keep_node(hkr3, uanf, left_comp, right_comp, ra,
              seg_start, seg_end, (char *)(ra->semant));
  }
}   /* end make_node */

// -----------------------------------------------------------
/* Analyseliste auf Terminal und in Ausgabefile ausgeben */
/* output the internal analysis list in a readable form  */
void print_analysis_list(NODE *start_nodes)
{
NODE *pnode;
int l, rnumber;
  /* Inputsatz ausgeben */
  fprintf(stdout, "%s\n",sentence);

  /* Kopf der Analyseliste ausgeben */
  printf("      Kat.                 Rnu.  LK   RK #   UKW\n");
  printf("-------------------------------------------------\n");

  /* vollstaendige Analyseliste ausgeben */
  pnode = start_nodes;
  while (pnode != NULL)
  {
    printf("%p ", pnode);
    for (l=0; l < LAHK; l++)
    {
      fputc(pnode->pkar->cstring[l], stdout);
    }
    rnumber = -1;
    if (pnode->prul != NULL) rnumber = pnode->prul->number;
    printf(" %p %p %4d", pnode->plko, pnode->prko,rnumber);
    printf("     %2d - %2d # ", pnode->lseg, pnode->rseg);
    
    /* UK-Werte (ohne UK - Namen! ) ausgeben */
    print_subcategories(pnode->pukw);
    //
    pnode = pnode->next;
    fprintf(stdout, "\n");
  }   // end while (pnode != NULL)
}   /* end print_analysis_list() */

// -----------------------------------------------------------
/* finde rechtsangrenzenden Knoten */
/* aspu = Aufsetzpunkt             */
/* find right neighbour node       */
NODE *find_rneighbour(NODE *adr,
                      NODE *aspu)
{
int sen;
  sen = adr->rseg + 1;
  adr = aspu;
  while (adr != NULL)
  {
    if (adr->lseg == sen)
        return(adr);
    adr = adr->next;
  }
  return(NULL);
}   /* end find_rneighbour() */

// -----------------------------------------------------------
/* UK-Werte auf Schnitt pruefen                     */
/* Rueckgabe > 0 wenn Schnitt vorhanden             */
/* check matching conditions for subcategory values */
/* a non zero intersection value must result        */ 
int check_scv(RULE *ra,
              NODE *ka,
              int   part)
{ 
SCVALUE *hkr,*hadr;
int l, j;
  hkr = ra->rpart[--part].pscv;
  if (hkr == NULL) return(1);
  hadr = ka->pukw;
  if (hadr == NULL) return(1);
  /* Laengen vergleichen */
  if (hkr->length != hadr->length) return(0);
  j = 0;
  for (l = 0; l < hkr->length; l++)
    if ((hkr->values[l] & hadr->values[j++]) == 0) return(0);
  return(1);
}   /* end check_scv() */

// -----------------------------------------------------------
/* Uebereinstimmung von Kategorien pruefen          */
/* Rueckgabe > 0 wenn Uebereinstimmung              */
/* Check matching condition for categories          */
int check_category(RULE *ra,
                   NODE *ka,
                   int   part)
{
CFRAME *hkr,*hadr;
  hkr  = ra->rpart[part-1].pcfr;
  hadr = ka->pkar;
  if (my_strcmp(hkr->cstring,hadr->cstring,LAHK) == 0)
  {
    /* UK-Werte auf Schnitt pruefen */
    if (check_scv(ra, ka, part)) 
      return(1);
  }
  return(0);
}   /* end check_category */

// -----------------------------------------------------------
/* UK-Wert aus Knoten liefern */
/* return a subcategory value */
int get_scv(NODE *kno,
            int   nuk)
{
SCVALUE *hkno;
  hkno = kno->pukw;
  if (nuk == 0) 
    return(-1);
  else
    return(hkno->values[nuk - 1]);
}  /* end get_scv() */

// -----------------------------------------------------------
/* Bedingungen puefen                         */
/* Rueckgabe > 0 wenn Bedingung zutrifft      */
/* Ab Adresse hukstr wird neue Merkmalsbelegung */
/* abgelegt                                     */
int compare_conditions(RULE *ra,
                       NODE *kno1,
                       NODE *kno2,
                       SCVALUE *hukstr)
{
SCVALUE *hadr,*pbd;
CHAR cw,*scp;
int lw;
  /* Kopie des Regel-UK-Strings der rechten */
  /* Seite anlegen                          */
  hadr = ra->rpart[2].pscv;
  hukstr->length = 0;
  if (hadr != NULL)
  {
    hukstr->length = hadr->length;
    my_strncpy(hukstr->values,hadr->values,hadr->length);
  }

  /* Bedingung da ? */
  if ((pbd = ra->rupbd) == NULL) return(1);

  lw  = pbd->length;
  scp = pbd->values;
  while (lw-- > 1)
  {
    cw = get_scv(kno1, *scp++);
    /* Schnitt pruefen */
    if (kno2 != NULL)
    {
      lw--;
      cw &= get_scv(kno2,*scp++); 
      if (cw == 0) return(0);
    }
    /* Schnittwert in Zielwertregister schieben */
    if (*scp != 0)
      hukstr->values[*scp - 1] = cw;
    lw--;
    scp++;
  }   /* while lw-- > 0 */
  return(1);
}   /* end compare_conditions() */

// -----------------------------------------------------------
/* Anwendung der Regeln                */
/* Die Regeln muessen sortiert sein !  */
/* Apply the rules (the grammar). Note */
/* that the rules must be sorted       */
void apply_rules()
{
CFRAME *cfr;
NODE *ka;
RULE *ra;
CHAR anfc;
NODE *la;
NODE *paxiom;
  /* Schleife, bis keine Regel mehr anwendbar */
  while (1)
  {
    zanw = 0;
    ka   = start_nodes;
    /* Schleife ueber alle Knoten */
    /* ka = Knotenadresse         */
    /* ra = Regelandresse         */
    while (ka != NULL)
    {
      anfc = ka->pkar->cstring[0];
      ra = (RULE *)(rhsh[anfc - 'A']);;

      /* Schleife ueber alle Regeln           */
      /* Einstieg in Regeln ueber Hashtabelle */
      /* loop over all rules                  */
      while (ra != NULL)
      {
        /* Zu weit in Regeln ? */
        cfr = ra->rpart[0].pcfr;
        if (cfr->cstring[0] > anfc) break;
        if (trace)
        {
          printf("ka,ra,rnu= %p %p %2d\n", ka, ra, ra->number);
        }
        /* Linke Regelseite 1 pruefen */
        if (check_category(ra,ka,1))
        {
          left_comp = ka;
          /* seg_start = Segmentanfang */ 
          seg_start = left_comp->lseg;
          /* seg_end = Segmentende   */
          seg_end = left_comp->rseg;
          /* Einer- oder Zweierregel? */
          if (trace)
          {
            printf("\tleft_comp,seg_start,seg_end=%p %4d %4d\n",
                    left_comp, seg_start, seg_end);
          }
          right_comp = NULL;
          if (ra->rpart[1].pcfr == NULL)
          { 
             /* Einerregel */
             if (compare_conditions(ra,left_comp,NULL,hukstr))
                /* neuen Knoten anlegen */
                make_node(ra);
          }
          else
          {
            /* Linke Regelseite 1 trifft zu */
            /* Zweierregel liegt vor        */
            /* Linke Regelseite 2 pruefen   */
            la = start_nodes;
            /* Schleife ueber alle Knoten */
            while (la != NULL)
            {
              /* rechtsangrenzendes Segment finden */
              right_comp = find_rneighbour(left_comp, la);
              if (right_comp == NULL) break;
              if (trace)
              {
                printf("\t\tla,right_comp=%p %p\n", la, right_comp);
              }
              la = right_comp->next; 
  
              /* linke Regelseite 2 pruefen */
              if (check_category(ra, right_comp, 2))
              { 
                seg_end = right_comp->rseg;
                /* Bedingungen pruefen */
                if (compare_conditions(ra,left_comp,right_comp,hukstr))
                {
                   /* Neuen Knoten anlegen */
                   make_node(ra);
                }
              }   /* if check_category(ra,right_comp,2) */
            }   /* while la != NULL */
          }   /* else */
        }   /* if check_category(ra,ka,1) */

        /* auf naechste Regel */
        ra = ra->next;
      }   /* while ra != NULL */

      /* auf naechsten Knoten im Kateg.speicher */
      /* go to next node                        */
      ka = ka->next;
   }   /* while (1) */

   /* Abbruch, wenn keine Regel angewendet wurde */
   /* exit if no rule applied                    */ 
   if (zanw == 0) 
     break;
  }   /* while (1) */ 

  /* Analysebaeume zeichnen, ggf. Analyseliste ausgeben        */
  /* Draw a linear brackets notation                           */
  /* Draw a two dimensional analysis tree if desired           */ 
  paxiom = find_axiom_node(start_nodes, AXIOM);
  if (paxiom != NULL)
  {
    // the axiom has been found: a well formed tree or bracketed
    // structure may be drawn
    bracket_buf = br_addr = (CHAR *)malloc(1000);
    *br_addr++ = '<';
    no_drawn_nodes   = 0;
    draw_brackets(paxiom);
    *br_addr++ = '>';
    *br_addr   = 0;
    hup_message(MSG_BRACKNOT);
    fprintf(stdout," %s\n", bracket_buf); 
    if (mode & OPTSUBCAT)
      print_all_subcategories();
    if (mode & OPTTREE)
      draw_analysis_tree((char *)bracket_buf);
    if (mode & OPTANALIST)
      print_analysis_list(start_nodes);
    //
    free(bracket_buf);
  }
  else
  {
    hup_message(MSG_NOTANALYZED);
    print_analysis_list(start_nodes);
  }

  if (interactive)
    enter_return();
}   /* end apply_rules() */

// -----------------------------------------------------------
/* Hashtabelle initialisieren */
/* initialize a hash table    */
void init_hash(CHAR **hsh)
{
int n;
  for (n=0; n < LHSH; n++)
    hsh[n] = NULL;
}   /* end init_hash() */

// -----------------------------------------------------------
/* Suche in einem Woerterbuch              */
/* Rueckgabe: Zeiger auf Laengeneintrag    */
/* Dictionary lookup                       */
LEXENT *dict_lookup(CHAR   *swa,
                    LEXENT *hash[],
                    int     lan)
{
LEXENT  *diad;
int     first_letter;
char    buf[256];
  if (trace > 1)
  {
    strncpy((char *)buf, (char *)swa, lan);
    buf[lan] = 0;
    printf("dict_lookup(%s)\n", buf);
    fflush(stdout);
  }
  /* Adr. im WB aus Hashtabelle ermitteln */
  first_letter = *swa - 'A';
  diad = hash[first_letter];

  while (diad != NULL)
  {
    /* Laenge + Wortlaut pruefen */
    if (compare_word(diad, swa, lan) != NULL) 
    {
      break;
    }
    /* auf naechstes Wort im WB */
    diad = diad->next;
    /* Zu weit im WB ? */
    if (diad->swo[0] > *swa) 
    {
      diad = NULL;
      break;
    }
  }   /* while diad != NULL */
  if (trace > 1)
  {
    if (diad != NULL)
      printf("dict_lookup():OK=%s\n", buf);
    else
      printf("dict_lookup():ERROR=%s\n", buf);
    fflush(stdout);
  }
  return(diad);
}   /* end dict_lookup() */

// -----------------------------------------------------------
/* Fortsetzung von dict_lookup                          */
/* aspu = Aufsetzpunkt im WB (Aufsetzen dahinter)       */
/* continuation of dictionary lookup                    */
LEXENT *continue_lookup(LEXENT *aspu,
                        CHAR   *swa,
                        int     lan)
{
  aspu = aspu->next;
  if (aspu == NULL) 
    return(NULL);
  /* Zu weit in WB ? */
  if (aspu->swo[0] > *swa) 
    return(NULL);
  if (compare_word(aspu,swa,lan)) 
    return(aspu);
  return(NULL);  
}   /* end continue_lookup() */

// -----------------------------------------------------------
/* Einfache morphologische Analyse                     */
/* Rueckgabe: Segmentnummer des letzten anal. Segments */
/* simple morphological analysis                       */
int morph_analysis(CHAR   *wanf,
                   LEXENT *hsh[],
                   int     anzw,
                   int     l)
{
CHAR *eanf;
LEXENT *wba1,*wba2,*headr;
int le;
  le = 1;
  while (--l > 2)
  {
    eanf = wanf + l;
    headr = end_start;
    /* In Endungstabelle suchen */
    while (headr != NULL)
    {
      /* Laenge pruefen */
      if (headr->length != le)
      {
         headr = headr->next;
         continue;
      }
      /* Wortlaut pruefen */
      if (my_strcmp(eanf, headr->swo, le) == 0)
      {
        /* Stamm suchen im WB */
        wba1 = dict_lookup(wanf, hsh, l);
        if (wba1 != NULL) 
        {
          /* Endung ebenfalls im WB suchen */
          wba2 = dict_lookup(eanf, hsh, le);
          if (wba2 != NULL)
          {
            while (wba1 != NULL)
            {
              /* Stamm als Knoten eintragen */
              keep_node(wba1->cframe,wba1->scvalue,
                   NULL,NULL,NULL,anzw,anzw,"");
              wba1 = continue_lookup(wba1,wanf,l);
            }   /* while wba1 != NULL */
            /* Endung als Knoten eintragen */
            anzw++;
            while (wba2 != NULL)
            {
              keep_node(wba2->cframe,wba2->scvalue,
                   NULL,NULL,NULL,anzw,anzw,"");
              wba2 = continue_lookup(wba2,eanf,le);
            }   /* while wba2 > NULL */
            return(anzw);
          }   /* if wba2 > NULL */        
        }   /* if wba1 > NULL */
      }
      headr = headr->next;
    }   /* while headr != NULL */
    le++;
  }   /* while --l > 2 */
  return(0);
}   /* end morph_analysis() */

// -----------------------------------------------------------
/* Lexikonsuche: allen Woertern des Analysesatzes           */
/*  werden Kategorien zugeordnet                            */
/* Stoesst die morpholog. Analyse an, falls erforderlich    */
/* Rueckgabe: Nummer des aeussest rechten Segments          */ 
/*          : < 0 wenn keine Identifikation                 */
/* Ordnet Zahlen die Kategorie des WB-Eintrags "NULL" zu    */
/* Ordnet Strings die Kategorie des WB-Eintrags "STRING" zu */ 
/* lexical analysis - dictionary lookup                     */
int lex_analysis(CHAR *sadr)
{
LEXENT *wbad;
int nseg,wordlen;
CHAR *anf,hbuf[100];
  if (trace > 1)
  {
    fprintf(stdout, "in lex_analysis()\n");
    fflush(stdout);
  }
  wordlen =  0;
  anzw    =  0;
  anf     = NULL;
  while (*sadr != 0)
  {
    if (strchr((char *)w_end_chars, (char)*sadr) != NULL)
    {
      /* Wortende erkannt */
      if (wordlen > 0)
      { 
        anzw++;
        if ((*anf >= '0') && (*anf <= '9'))
        {
           /* Wort muss Zahl sein: Semantem ist Zahl selbst! */
           wbad = dict_lookup((CHAR *)"NULL", whsh, 4);
           if (wbad == NULL) 
             errexit(" Kategorie NULL nicht definiert ");
           my_strncpy(hbuf,anf,wordlen);
           hbuf[wordlen] = 0;
           keep_node(wbad->cframe, wbad->scvalue,
                     NULL, NULL, NULL,
                     anzw, anzw, (char *)hbuf);
        }
        else
        {
          /* Normales Wort: weder Zahl noch String */
          /* Im Woerterbuch suchen                 */
          wbad = dict_lookup(anf, whsh, wordlen);
          if (wbad != NULL)
          {
            while (wbad != NULL)
            {
              /* Knoten im Kat-Speicher ablegen */
              keep_node(wbad->cframe, wbad->scvalue,
                        NULL, NULL, NULL,
                        anzw, anzw, (char *)(wbad->semant));
              /* Naechsten Eintrag im WB pruefen */
              wbad = continue_lookup(wbad, anf, wordlen);
            }   /* while wbad > 0 */
          }   /* if wbad != NULL */
          else 
          {
            /* Kein vollstaendiger Eintrag im WB da */
            /* Versuche Morpholog. Analyse       */
            if ((nseg = morph_analysis(anf, whsh, anzw, wordlen)) == 0)
            {
              // Der Eintrag wurde nicht identifiziert
              // The entry could not be identified
              make_string(hbuf, anf, wordlen);
              hup_message(MSG_LEXNOTFOUND);
              fprintf(stdout," %s\n", hbuf);
              if (interactive)
                enter_return();
              return -1;
            }
            else anzw = nseg;
          }   /* else */
        }   /* else */
      }   /* if wordlen > 0 */
      wordlen = 0;
    }  /* if strchr() .... */
    else
    {
      if (++wordlen == 1)
        /* Wortanfang merken */
        anf = sadr;
    }   /* else  */
    sadr++;
  }  /* while *sadr != 0 */
  //
  return(anzw);
}   /* end lex_analysis */

// -----------------------------------------------------------
/* Naechsten Inputsatz einlesen              */
/* Wird in Majuskeln in dest_addr abgelegt   */
/* Rueckgabe: Anfangsadresse des Satzes      */
/*            NULL bei EOF                   */
/* Read next input sentence                  */
/* Returns: NULL on EOF, else dest_addr      */
/* The sentence is converted to upper case   */
CHAR *read_sentence(FILE *inp,
                    CHAR *dest_addr,
                    int ldest)
{
int n, c, len;
  memset(dest_addr, 0, ldest);  // 07-2002
  //dest_addr[0] = 0;
  if (inp == stdin)
  {
    // read the sentence from stdin
    hup_message(MSG_NEXTSENTENCE);
    if (fgets((char *)dest_addr, ldest, stdin) == NULL)
      return NULL;
  }
  else
  {
    // read the sentence from a file
    if (fgets((char *)dest_addr, ldest, inp) == NULL)
      return NULL;
  }
  len = 0;
  for (n=0; dest_addr[n] != 0; n++)
  {
     c = dest_addr[n];
     if ((c & 0xff) < ' ')
     {
       // eliminate CR, LF etc. 
       dest_addr[n] = 0;
       continue;
     }
     len++;
     dest_addr[n] = my_toupper((CHAR)c);
     /* Elisionszeichen (Apostroph) */
     /* if (c == 39) dest_addr[n] = ' ';  */   /* 1.1.89 E.H. */
     /* Satzende ? */
     if (strchr((char *)s_end_chars, (char)c) != NULL) 
       dest_addr[n] = 0;
   }
   if (len > 0) 
   {
     strcat((char *)dest_addr," ");
     if (inp != stdin)
     {
       fprintf(stdout, "Satz: %s\n", dest_addr);
       fflush(stdout);
     }
     no_sentences++;
     return dest_addr;
   }
   else
     return NULL;
}   /* end read_sentence() */

// -----------------------------------------------------------
/* Hauptkategorie in Verkettung ablegen                  */
/* nicht ablegen, falls schon existent                   */
/* generate a main category pattern in the linked list   */
CFRAME *put_category(CHAR *kr)
{
CFRAME *kar;
int lkr;
  /* Kategorie schon da? */
  kar = start_categories;
  while (kar != NULL)
  { 
    if (my_strcmp(kar->cstring,kr,kar->length)== 0) return(kar);
    kar = kar->next;
  }
  /* Kategorie ist neu */
  lkr = strlen((char *)kr);
  kar = (CFRAME *)malloc(sizeof(CFRAME) + lkr);
  if (kar == NULL)
     errexit("put_category: malloc Fehler");
  if (start_categories == NULL) start_categories = kar;
  /* Laenge  ablegen */
  kar->length = lkr;
  /* Nachfolgepointer ablegen */
  kar->next = NULL;
  if (last_category != NULL) last_category->next = kar;
  /* Kategorienstring ablegen */
  my_strncpy(kar->cstring,kr,lkr);
  last_category = kar;
  return(kar);
}  /* end put_category */

// -----------------------------------------------------------
/* Bit (Pos. als ASCII Zeichen in wert) in Byte adr setzen */
void setbit(CHAR *adr,
            CHAR  wert)
{
int w;
  if (wert == '0') 
    return;
  w = 1;
  wert -= '1';
  if (wert > 0) *adr += (w << wert);
  else *adr = *adr + w;
}   /* end setbit */

// -----------------------------------------------------------
/* UK-Werte komprimieren und ablegen             */
/* Der Wortlaut der UKs wird nicht abgelegt !    */
/* Die UK-Werte sind einmalig pro Knoten         */
/* Sie bilden deshalb keine eigene Verkettung    */
/* Store a subcategory pattern value             */
/* the names of the subcategories are not stored */ 
SCVALUE *put_scv(CHAR *uk)
{
CHAR puf[20],lc,c;
int luk;
SCVALUE *ukw;
  if (*uk == 0) 
    return(NULL);
  lc = 0;
  luk = -1;

  /* Nur die Werte extrahieren und in einem */
  /* Byte pro UK zusammenfassen             */
  while (*uk != 0)
  {
    c = *uk++;
    if (c == ';') break;
    if (c == ')')
    {
      /* () bedeutet: alle Werte zugelassen */
      if (lc == '(') puf[luk] = 0xff;
    }  
    if (c == '(')
      puf[++luk] = 0;
    if (strchr("12345678", c) != NULL)
    {
       /* UK-Wert muss einstellig sein und im */
       /* Bereich 1-8 liegen                  */
       /* entspr. Bit setzen                  */
       setbit(&puf[luk], c);
    }
    lc = c;
  }  /* while *uk != 0 */
  ukw = (SCVALUE *)malloc(sizeof(SCVALUE)  +  ++luk );
  if (ukw == NULL)
    errexit("put_scv: malloc Fehler");
  ukw->length = luk ;
  my_strncpy(ukw->values,puf,luk);
  return(ukw);
}  /* end put_scv */

// -----------------------------------------------------------
/* Bedingungen ablegen: Bedingungen werden wie  */
/* Werte abgelegt                               */
/* Store subcategory match conditions           */
SCVALUE *put_condition(CHAR *str)
{
int l;
CHAR hpuf[20];
SCVALUE *pbd;
  if (*str == 0) 
    return(NULL);
  l = 0;
  while (*str != 0)
  {
    if (strchr("0123456789", *str) != NULL)
      hpuf[l++] = *str - '0';
    str++;
  }   /* while *str != 0 */
  pbd = (SCVALUE *)malloc(sizeof(SCVALUE) + l);
  if (pbd == NULL)
    errexit("put_condition: malloc Fehler");
  /* Laenge ueber alles ablegen */
  pbd->length = l;
  /* Wertestring ablegen */
  my_strncpy(pbd->values,hpuf,l);
  return(pbd);
}   /* end put_condition */

// -----------------------------------------------------------
/* Woerterbuch einlesen  - Eintraege im Speicher ablegen  */
/* Hashtabelle erstellen                                  */
/* Rueckgabe: Adresse des ersten Eintrags                 */
/* Hier findet keine Syntaxpruefung statt: die Eintraege  */
/* muessen syntaktisch korrekt sein                       */  
/* Read the lexicon, built up the hash table              */
/* no syntax check for lexicon here                       */
LEXENT *read_dict(FILE   *fil,
                  LEXENT *hash[])
{
CHAR c,ll,kat,sword[100],fe[100];
CHAR kr[100],uk[100];
CHAR *akr,*auk,lka,*asem;
LEXENT *first,*hadr;
int ic,psw,lsem;

  ll    = 0;
  kat   = 0;
  llast = NULL;
  first = NULL;
  akr   = kr;
  auk   = uk;
  *auk  = 0;
  asem  = sembuf;
  *asem = 0; 
  psw   = 0;
  lsem  = 0;
  lka   = 0;
  
  while (1)
  {
    c = ic = fgetc(fil);
    if (ic == EOF) 
      break;
    if (c == CR) 
      continue;
    if ((c == '$') && (lsem == 0))
    {
      /* Kommentar: Rest ueberlesen */
      fgets((char *)fe, sizeof(fe), fil);
      continue;
    }
    if (trace) 
      fputc(c,stdout);
    if (c == LF)
    {
      /* Ende des Eintrags */
      hadr = (LEXENT *)malloc(sizeof(LEXENT) + psw);
      if (hadr == NULL)
        errexit("read_dict: malloc Fehler");
      if (first == NULL) first = hadr;

      /* Suchwort ablegen */
      my_strncpy(hadr->swo,sword,psw);     
 
      /* Laenge Suchwort */
      hadr->length = psw;

      /* Hashadresse ablegen */
      ll = sword[0] - 'A';
      if (hash[ll] == NULL) 
        hash[ll] = hadr;     
 
      /* Kategorialrahmen ablegen */
      hadr->cframe = put_category(kr);

      /* Pointer UK-Werte ablegen */
      hadr->scvalue = put_scv(uk);

      /* Semantem (= String)  ablegen */
      strcpy((char *)hadr->semant, (char *)sembuf);

      /* Pointer auf Nachfolger  im Vorgaenger setzen */
      hadr->next = NULL;
      if (llast != NULL) 
        llast->next = hadr;
      llast = hadr;

      auk = uk;
      akr = kr;
      *auk = 0;
      asem      = sembuf;
      sembuf[0] = 0;
      kat = 0; 
      psw = 0;
      lsem = 0;
    }
    else
    {
      if (kat == 1)
      /* Kategorie und UKs */
      {
        if (c == ':')
        {
           kat  = 0;
           lsem = 1;
           continue;
         }
        lka++;
        if (lka <= LAHK)
        { 
          *akr++ = c;
          *akr   = 0;
        }
        if (lka > LAHK)
        {
          *auk++ = c;
          *auk   = 0;
        }
      }  /* if kat == 1 */

      if (c == DICT_SEP) 
      {
        /* Trenner zur Kategorie */
        kat = 1;
        lka = 0;
      }

      if (lsem)
      {
        if ((c == ' ')&& (sembuf[0] == 0)) continue;
        if (lsem++ >= (LSEM - 1)) continue;
        *asem++ = c;
        *asem   = 0;
        continue;
      }

      if (kat == 0)
      { 
         /* Suchwort in Majuskeln ablegen */
         sword[psw++] = my_toupper(c);
         sword[psw]   = 0;
      }
    }  /* else */
  }  /* while 1 */
  //
  return(first);
}   /* end read_dict() */

// -----------------------------------------------------------
/* Anfang-Ende eines Regelteils                 */
/* globale Variablen: anzel,rul_start,last_rule */
void store_rule(CHAR *kr,
                CHAR *uk)
{
int anfc, n;
  if (anzel == 0)
  {
    last_rule = (RULE *)malloc(sizeof(RULE));
    if (last_rule == NULL)
      errexit("store_rule: malloc Fehler");      
    if (rul_start == NULL) rul_start = last_rule;
    /* Regel vorbesetzen */
    last_rule->next = NULL;
    last_rule->rupbd = NULL;
    for (n=0; n < 3; n++)
    {
      last_rule->rpart[n].pcfr  = NULL;
      last_rule->rpart[n].pscv = NULL;
    }
    /* Hashtabelle rhsh besetzen */
    anfc = *kr - 'A';
    if (rhsh[anfc] == NULL) rhsh[anfc] = (CHAR *)last_rule;
  }
  /* Pointer auf Kategorie ablegen */
  last_rule->rpart[anzel].pcfr = put_category(kr);
  /* Pointer auf UK-Werte ablegen */
  last_rule->rpart[anzel].pscv = put_scv(uk);
  anzel++;
  *kr = 0;
  *uk = 0;
}   /* end store_rule() */

// -----------------------------------------------------------
/* Grammatik(Regeln) lesen                     */
/* Regeln muessen sortiert sein !              */
/* read a grammar - the rules must be sorted   */
/* No syntax check here!                       */
/* Use the external program syncheck() instead */
void read_grammar(FILE *kfil)
{
CHAR c, *akr, *auk;
int  renu, ic, l;
CHAR kr[80], uk[80];
CHAR fe[100];
FILE *rfil;
  
  if (fgets((char *)kr ,sizeof(kr), kfil) == NULL)
  {
    errexit("no entry for grammar file in the confi.file");
  }
  if (trace)
    fprintf(stdout, "read_grammar(): file=%s\n", kr);
  kr[strlen((char *)kr) - 1] = 0;
  rfil = fopen((char *)kr, "r");
  if (rfil == NULL) 
  {
    if (trace)
      fprintf(stdout, "read_grammar(): error 1: file=%s\n", kr);
    errexit("Regeln: Dateifehler");
  }
  
  /* Lesen und Ablegen der Regeln           */
  /* Hashtabelle rhsh fuer Regeln erstellen */
  /* Kategorien und UK-Werte separat        */
  init_hash((CHAR **)rhsh);
  if (trace) printf("\n");
  renu = 0;

  while (1)
  {
    akr   = kr;
    auk   = uk;
    *auk  = 0;
    *akr  = 0;
    anzel = 0;
    renu++;
    l     = 0;
    c     = 0;
    lsem  = 0;

    while (c != '.')
    {
      c = ic = fgetc(rfil);
      if (ic == EOF) break;
      /* Kommentare ueberlesen */
      if ((c == '$') && (lsem == 0))
      {
        fgets((char *)fe, sizeof(fe), rfil);
        continue;
      }
      if (c == LF) continue;
      if (c == CR) continue;
      if (trace)
      {
        if ((anzel == 0) && (kr[0] == 0)) printf("\n%4d ",renu); 
        fputc((char)c, stdout);
      }
      if ((c == ' ') && (lsem <= 0)) continue;
      l++;
      /* Trenner zweiter Teil linke Seite ? */
      if (c == '+')
      {
        store_rule(kr, uk);
        if (trace) printf("\n\t");
      }
      /* Trenner rechte Regelseite ? */
      if ((c == '=') && (lsem == 0)) 
      { 
        store_rule(kr, uk);
        /* Einerregel ? */
        if (anzel == 1)
        {
          last_rule->rpart[1].pcfr = NULL;
          last_rule->rpart[1].pscv = NULL;
          anzel++;
        }
        if (trace) 
          printf("\n\t");
      }
      /* Trenner Bedingungsteil ? */
      if (c == ';')
      {
        store_rule(kr, uk);
        if (trace) 
          printf("\n\t");
      }
      /* Regelende ? */
      if (c == '.')
      {
        last_rule->rupbd  = put_condition(uk);
        last_rule->number = renu;
        if (lsem > 0)
        {
           /* Blank am Semantemende vermeiden */
           lsem = strlen((char *)sembuf);
           if (sembuf[lsem-1] == ' ') sembuf[lsem-1] = 0;
        }
        strcpy((char *)(last_rule->semant), (char *)sembuf);
        if (trace) 
           printf(" --- %p", last_rule);
        break;
      }
      /* Beginn Semantem? */
      if (c == ':')
      {
        lsem      = 1;
        sembuf[0] = 0;
        continue;
      }  
      if (lsem) 
      {
        if ((c == ' ') && (sembuf[0] == 0)) continue;
        /* In Semantemstring */
        if (lsem >= LSEM) continue; 
        sembuf[lsem-1] = c;
        sembuf[lsem++] = 0;
        continue;
      }
      else
      { 
        /* Zeichen im Kategorialrahmen ablegen */
        if (l <= LAHK)
        {
          *akr++ = c;
          *akr   = 0;
        }
        if (l > LAHK)
        {
          *auk++ = c;
          *auk   = 0;
        }
      }   /* else */
      /* Adressen Kateg.rahmen zuruecksetzen */
      if (strchr("+=;:", c) != NULL)
      {
        akr = kr;
        auk = uk;
        if (c != ';') l = 0;
        lsem = 0;
        sembuf[0] = 0;
      }
    }   /* while c != '.' */

     if (ic == EOF) break;
     /* Nachfolgepointer setzen */
     last_rule->next = NULL;
     if (prev_rule != NULL) 
       prev_rule->next = last_rule;
     prev_rule = last_rule;
   }   /* while (1) */
   fclose(rfil);
}   /* end read_grammar */ 

// -----------------------------------------------------------
/* Naechstes Item aus iz in rz ablegen */
int next_item(CHAR *iz, 
              CHAR *rz)
{
int l;
  l = 0;
  while (*iz != 0)
  {
    if (*iz == ',') break;
    if (*iz == CR) break;
    if (*iz == LF) break;
    l++;
    *rz++ = *iz++;
  }
  *rz = 0;
  if (*iz == ',') l++;
  return(l);
}   /* end  next_item */

// -----------------------------------------------------------
/* Zeile  mit Codesequenz dekodieren */
void decodl(CHAR *zeile,
            CHAR *str)
{
int i,l;
CHAR *hadr,hstr[20];
  hadr = zeile;
  while (1)
  {
    l = next_item(hadr,hstr);
    sscanf((char *)hstr, "%d",  &i);
    *str++ = i;
    hadr += l;
    if (*hadr == 0) break;
    if (*hadr == CR) break;
    if (*hadr == LF) break;
  }
  *str = 0;
}   /* end decodl */

// -----------------------------------------------------------
/* read the message file and keep the messages in memory      */
/* messages must have the form: nn string                     */
/* where nn is a decimal message number followed by a blank   */
/* the message string may contain format characters following */
/* the standard C syntax                                      */ 
MESSAGE *read_messages(FILE *kfil)
{
  FILE *mf;
  MESSAGE *mact,*mstart,*mlast;
  CHAR hbuf[128];
  int iw,l,ll,mnumber;
  //
  if (fgets((char *)hbuf, sizeof(hbuf), kfil) == NULL) 
    return(NULL);
  /* eliminate LF in file name */
  hbuf[strlen((char *)hbuf)-1] = 0;
  mf = fopen((char *)hbuf, "r");
  if (mf == NULL) 
    return(NULL);
  mstart = mlast = NULL;

  while (1)
  {
    if (fgets((char *)hbuf,sizeof(hbuf),mf) == NULL) 
      break;
    if (hbuf[0] == '\n') 
      continue;
    if (trace) 
      fprintf(stdout,"%s",hbuf);
    hbuf[strlen((char *)hbuf)-1] = 0;
    l = 0;
    for (l=0; hbuf[l] != 0; l++)
    {
      if (hbuf[l] == ' ')
      {
        /* bank found: separates number from string */
        iw = sscanf((char *)hbuf, "%d", &mnumber);
        if (iw <= 0) 
          return(NULL);
        ll = strlen((char *)(&hbuf[l+1]));
        break;
      } 
    }
    if (l <= 0) 
      return(NULL);
    mact = (MESSAGE *)malloc(sizeof(MESSAGE) + ll + 1);
    if (mact == NULL) 
      return(NULL);
    strcpy((char *)(mact->string), (char *)(&hbuf[l+1]));
    mact->number = mnumber;    
    mact->next   = NULL;
    if (mstart == NULL) mstart      = mact;
    else                mlast->next = mact;
    mlast = mact;
  }
  //
  fclose(mf);
  return(mstart);  
}   /* end read_messages */

// -----------------------------------------------------------
// check the memory: test how many blocks we can allocate
// This has no functional meaning - it is only used
// to detect memory losses
void check_memory(int release)
{
long size;
char *caddr;
char *array[2048];
int  n, count;
  size = 1024L * 1024L;
  count = 0;
  for (n=0; n < 2048; n++)
  {
    caddr = (char *)malloc(size);
    if (caddr == NULL) break;
    array[count++] = caddr;
  }
  printf("%d Blcke Memory a 1 MB allokiert\n", count);
  fflush(stdout);
  if (release)
  {
    for (n=0; n < count; n++)
    {
      if (array[n] == NULL) break;
      free(array[n]);
    }
  }
}   // end check_memory()  

// -----------------------------------------------------------
// Read the config file
// Read the lexicon and keep it in memory
// Read the rules and keep them in memory
void initialize_everything(char *config_fn,
                           int  lconfig_fn)
{
int  n;
FILE *kfil;
FILE *rfil;
char buf[256];
  if (trace)
  {
    fprintf(stdout, "in initialize_everything()\n");
    fflush(stdout);
  }
  //
  /* Konfig.File oeffnen */
  // Open the configuration file
  if (strlen(config_fn) == 0)
  {
    fprintf(stdout,"\nName configuration file:");
    if (fgets(config_fn, lconfig_fn, stdin) == NULL)
      errexit("No configuration file");
  }
  for (n=0; config_fn[n] != 0; n++)
  {
    if ((config_fn[n] & 0xff) < ' ')
      config_fn[n] = 0;
  }
  kfil = fopen(config_fn, "r");
  if (kfil == NULL) 
    errexit("Konfig.: Dateifehler");
  if (trace)
  {
    fprintf(stdout, "in initialize_everything(): config.file %s read\n",
            config_fn);
    fflush(stdout);
  }
  
  // read config. file, lexicon, rules
  /* read list of messages */
  start_messages = read_messages(kfil);
  if (start_messages == NULL) 
    errexit("Message file");
  if (trace)
  {
    fprintf(stdout, "in initialize_everything(): messages read\n");
    fflush(stdout);
  }

  /* Liste der Wortendezeichen lesen */
  // Read the list of word ending character codes
  if (fgets(buf,sizeof(buf),kfil) == NULL) 
     errexit("Error config. file: word end chars");
  decodl((CHAR *)buf, w_end_chars);
  if (trace)
  {
    fprintf(stdout, "in initialize_everything(): word delimiters read\n");
    fflush(stdout);
  }

  /* Liste der Satzendezeichen */
  // Read the list of sentence ending character codes
  if (fgets(buf, sizeof(buf), kfil) == NULL) 
    errexit("Error config.file: sentence delimiter chars");
  decodl((CHAR *)buf, s_end_chars);
  if (trace)
  {
    fprintf(stdout, "in initialize_everything(): sentence delimiters read\n");
    fflush(stdout);
  }

  /* Grammatik einlesen. Setzt Regelanfangsadr. rul_start */
  // Read the grammar (the grammatical rules)
  read_grammar(kfil);
  hup_message(MSG_GREAD);
  if (trace)
  {
    fprintf(stdout, "in initialize_everything(): grammar read\n");
    fflush(stdout);
  }

  /* Lexikon einlesen. Setzt Lexikonanfangsadr. lex_start */
  // Read the lexicon (the lexical rules)
  if (fgets(buf, sizeof(buf), kfil) == NULL) 
    errexit("Konf.Fehler");
  buf[strlen(buf) - 1] = 0;
  rfil = fopen(buf,"r");
  if (rfil == NULL) 
    errexit("lexicon file error (not found)");
  init_hash((CHAR **)whsh);
  lex_start = read_dict(rfil, whsh);
  hup_message(MSG_LREAD);
  fclose(rfil);

  /* Endungen lesen */
  // Read the list of endings (word morphology)
  if (fgets(buf, sizeof(buf), kfil) == NULL) 
    errexit("Konf.Fehler");
  buf[strlen(buf) - 1] = 0;
  rfil = fopen(buf, "r");
  if (rfil == NULL) 
    errexit("error endings file (not found)");
  init_hash((CHAR **)ehsh);
  end_start = read_dict(rfil, ehsh);
  hup_message(MSG_EREAD);
  fclose(rfil);
}   // end read_everything()

void usage()
{
  fprintf(stdout,"\nHUpars 09-2002 Copyright Dr. Edgar Huckert");
  fprintf(stdout,"\nusage: HUpars [-c configfile] [-tree] [-list]");
  fprintf(stdout,"\n              [-subcat] [-i inptext] [-a axiom] [-v]\n");
}   // end usage()

// -----------------------------------------------------------// -----------------------------------------------------------
/* Aufruf: kfgr  [-c configfile -m -t -a axiom]               */
/* -c : configuration file name follows                       */
/* -m : Modus  folgt                                          */
/* -t : trace on                                              */
/* -a : anderes Axiom als SATZ folgt                          */    
int main(int argc, char *argv[])
{
int  l;
char config_fn[128];
char inp_text[128];
FILE *inptxt;
long t1, t2;

  //check_memory(0);
  usage();
  //
  strcpy((char *)AXIOM, "SATZ");   // Standard - Axiom vorbesetzen */
  strcpy(config_fn, "");
  strcpy(inp_text, "");
  inptxt      = NULL;
  trace       = 0;
  interactive = 1;
  mode        = OPTBRACKETS;

  /* uebergebene Argumente auswerten */
  // Analyze command line arguments
  for (l = 1; l < argc; l++)
  {
    /* extract name of configuration file */
    if (strcmp(argv[l],"-c") == 0)
       strcpy(config_fn,argv[l+1]);
    /* set trace option */
    if (strcmp(argv[l],"-t") == 0) trace = 1;
    if (strcmp(argv[l],"-v") == 0) trace = 1;
    /* set new axiom string */ 
    if (strcmp(argv[l],"-a") == 0)
        strcpy((char *)AXIOM, argv[++l]);
    if (strcmp(argv[l],"-tree") == 0)
        mode |= OPTTREE;
    if (strcmp(argv[l],"-list") == 0)
        mode |= OPTANALIST;
    if (strcmp(argv[l],"-subcat") == 0)
        mode |= OPTSUBCAT;
    if (strcmp(argv[l],"-i") == 0)
      strcpy(inp_text, argv[l+1]); 
  }

  start_categories      = NULL;
  last_category         = NULL;
  rul_start             = NULL;
  last_rule             = NULL;
  prev_rule             = NULL;
  lex_start             = NULL;
  end_start             = NULL;

  // read config. file, lexicon, rules
  initialize_everything(config_fn, sizeof(config_fn));

  // open the input file (if specified)
  if (strlen(inp_text) > 0)
  {
    interactive = 0;
    inptxt = fopen(inp_text, "r");
    if (inptxt == NULL)
    {
      errexit("Fehler: Inputtext-Datei nicht gefunden");
    }
  }
 
  hukstr = (SCVALUE *)malloc(sizeof(SCVALUE) + 20);
  
  /* Hauptschleife */
  // Enter the main loop
  if (interactive)
    enter_return();
  t1 = get_epoch_time_ms();
  
  while (1)
  {
   /* naechsten Satz lesen und an  */
   /* Adresse sentence bereitstellen  */
   // Read the next sentence and store it at address sentence
   if (interactive)
   {
     sadr = read_sentence(stdin, sentence, sizeof(sentence));
   }
   else
   {
     sadr = read_sentence(inptxt, sentence, sizeof(sentence));
   }
   if (sadr == NULL)  
     break;

   /* Lexikonsuche und Wortmorphologie */
   // Lexicon search and morphology
   start_nodes = NULL;
   maxseg      = lex_analysis(sadr);
 
   /* Grammatische Regeln anwenden */
   // Apply the grammatical rules
   if (maxseg > 0) 
   {
     sembuf[0] = 0; 
     apply_rules();
   }
  }   /* while (1) */

  if (inptxt != NULL)
    fclose(inptxt);

  t2 = get_epoch_time_ms();
  //if (trace)
  fprintf(stdout, "\nZeit:%ld (ms) fuer %d Saetze\n", 
         t2 - t1, no_sentences);

  //check_memory(1);
  errexit("Ende");
  return 0;
}   /* end main */
